---
jupytext:
  formats: md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
    jupytext_version: 1.14.1
kernelspec:
  display_name: Python 3 (ipykernel)
  language: python
  name: python3
---

# Bag of Words erstellen

* J.Busse, www.jbusse.de, 2023-04-20, 2023-04-27

Lizenz: public domain / [CC 0](https://creativecommons.org/publicdomain/zero/1.0/deed.de) 

Dateien:
* {download}`regex_XX.zip`
* auspacken; das Verzeichnis `regex_XX/` sollte idealerweise als Geschwister-Verzeichnis des aktuellen Arbeitsverzeichnis `md/` angelegt werden -- oder ggf. Variable `path_to_md` anpassen.

```{code-cell} ipython3
import numpy as np
import pandas as pd
```

### Global Parameters

```{code-cell} ipython3
# path to files, incl. glob mask
#path_to_md = "md/*.md"
path_to_md = "../regex_XX/*.md"
```

```{code-cell} ipython3
# show intermediary results
# 0 none, 1 informative, 2 debug
verbosity = 1

def verbose(level,item):
    if level <= verbosity:
        display(item)
```

Read Files 
----

```{code-cell} ipython3
# https://stackoverflow.com/questions/3207219/how-do-i-list-all-files-of-a-directory
import glob
files = glob.glob(path_to_md)
verbose(1,f"{len(files)} Dateien gefunden")
# files
```

Wir lesen die Daten in das Dictionary `corpus_dict` ein:

```{code-cell} ipython3
def clean_string(string):
    """Liefert string "gesäubert" wieder als String zurück: 
    Interpunktionen, Sonderzeichen etc. 
    werden wie Leerzeichen behandelt."""
    alnum = lambda x: x if x.isalnum() else " "
    return "".join(alnum(c) for c in string )
```

```{code-cell} ipython3
corpus_dict_of_strings = {}

for file in files:
    with open(file, 'r') as f:
        corpus_dict_of_strings[file] = f.read()
# corpus_dict_of_strings
```

```{code-cell} ipython3
corpus_dict_of_lists = { doc: clean_string(text).split() 
                        for doc, text in corpus_dict_of_strings.items() }
# corpus_dict_of_lists
```

```{code-cell} ipython3
doclen = { doc: len(wordlist) for doc, wordlist in corpus_dict_of_lists.items()}
doclen
```

`all_words_set`: Alle Wörter, die im "gesäuberten" Korpus vorkommen

```{code-cell} ipython3
all_words_set = { w for d in corpus_dict_of_lists.values() for w in set(d) }
" ".join(sorted(list(all_words_set)))[:80]
len(all_words_set)
```

`all_words_dict`: Ein Dictionary, in dem wir zählen, wie häufig ein Wort insgesamt im Korpus vorkommt. Um später Speicherplatz und Laufzeit zu sparen, gehen wir den Korpus einmal komplett durch und zählen.

```{code-cell} ipython3
# Initialisierung: dict anlegen und für jedes Wort auf die Häufigkeit 0 setzen
all_words_count = { k: 0 for k in all_words_set}

for text in corpus_dict_of_lists.values():
    for word in text:
        all_words_count[word] += 1
# all_words_count
```

Für die Textähnlichkeit interessieren nur Wörter, die nicht nur in einem einzigen Dokument vorkommen. ... Und wir filtern auch gleich bedeutungslose Wörter heraus.

```{code-cell} ipython3
# Quelle: https://www.ranks.nl/stopwords/german
stoppwoerter = """aber als am an auch auf aus bei bin bis bist 
da dadurch daher darum das daß dass dein deine dem den der des 
dessen deshalb die dies dieser dieses doch dort du durch ein 
eine einem einen einer eines er es euer eure für hatte hatten 
hattest hattet hier hinter ich ihr ihre im in ist ja jede jedem 
jeden jeder jedes jener jenes jetzt kann kannst können könnt 
machen mein meine mit muß mußt musst müssen müßt nach nachdem 
nein nicht nun oder seid sein seine sich sie sind soll sollen 
sollst sollt sonst soweit sowie und unser unsere unter vom von 
vor wann warum was weiter weitere wenn wer werde werden werdet 
weshalb wie wieder wieso wir wird wirst wo woher wohin zu 
zum zur über""".split()

len(stoppwoerter)
```

```{code-cell} ipython3
early_word_elimination = True
if early_word_elimination:
    all_words_count_nosingles = { word: count for word, count in all_words_count.items() 
                                 if count > 1 and not word.lower() in stoppwoerter }
else:
    all_words_count_nosingles  = all_words_count
# print(all_words_count_nosingles)
len(all_words_count_nosingles)
```

Welche Wörter haben wir verworfen, weil sie nur einmal vorkommen oder Stoppwörter sind? Wie viele sind das?

```{code-cell} ipython3
verworfen = set(all_words_count) - set(all_words_count_nosingles)
# print(nureinmal)
# verworfen
```

`document_word_count`: Ein Dictionary von Dictionaries: Für jedes Dokument zählen wir für jedes Wort in dem Dokument, wie häufig das Wort im Dokument vorkommt.

```{code-cell} ipython3
document_word_count = {}

verbosity = 1

# gehe durch unseren Korpus durch
for doc, text in corpus_dict_of_lists.items():
    
    verbose(2, text) # debug
    
    # Initialisiere für das aktuelle Dokument ein neues, leeres Dictionary
    # word count absolut
    document_word_count[doc] = dict.fromkeys(all_words_count_nosingles, 0)
    
    # für jedes Wort im aktuellen Dokument
    for word in text:
        
        # kommt das aktuelle Wort öfters als nur einmal im Korpus vor?
        if word in all_words_count_nosingles:
            document_word_count[doc][word] += 1

# document_word_count
```

```{code-cell} ipython3
# total_word_count_from_document_word_count
twcfdwc  = dict.fromkeys(all_words_count_nosingles, 0)

for doc, text in corpus_dict_of_lists.items():
    for word in text:
        if word in twcfdwc:
            twcfdwc[word] += 1
# twcfdwc
```

```{code-cell} ipython3
if not early_word_elimination:
    twcfdwc_nosingles = { word: count for word, count in twcfdwc.items() 
                                 if count > 1 and not word.lower() in stoppwoerter }
else:
    twcfdwc_nosingles = twcfdwc
twcfdwc_nosingles
```

## Exkurs:  des Winkels zwischen zwei Vektoren

https://www.gut-erklaert.de/mathematik/skalarprodukt-winkel-vektoren-formel.html

```{code-cell} ipython3
import numpy as np
```

```{code-cell} ipython3
a = np.array((1, 2, 3))
b = np.array((4, 5, 6))
```

```{code-cell} ipython3
np.dot(a, b)
```

```{code-cell} ipython3
cos_alpha = np.dot(a, b) / ( np.dot(a,a)**.5 * np.dot(b,b)**.5 )
cos_alpha
```

## Cosinus zwischen zwei unserer Texte

+++

Nochmal zur Erinnerung: die Liste `files` enthält die Dateinamen unseres Korpus:

```{code-cell} ipython3
files[0]
```

Also können wir so auf die Wörterzählungen der ersten Datei zugreifen, wenn wir den Dateinamen nicht tippen wollen:

```{code-cell} ipython3
print(list(document_word_count[files[0]].values()))
```

```{code-cell} ipython3
a = np.array(list(document_word_count['../regex_XX/regex_8.md'].values())) # erste Datei, hier '../regex_XX/regex_8.md'
b = np.array(list(document_word_count[files[1]].values())) # zweite Datei, hier '../regex_XX/regex_6.md'
```

```{code-cell} ipython3
cos_alpha = np.dot(a, b) / ( np.dot(a,a)**.5 * np.dot(b,b)**.5 )
cos_alpha
```

## Matrix erstellen

"Zu Fuß", noch im Lernprozess: Wir müssen unser dict `document_word_count` in eine "Matrix" umwandeln, hier zunächst eine Liste von Listen.

```{code-cell} ipython3
matrix = [ list(document_word_count[f].values()) for f in files ]
print(matrix)
```

Um Bibliotheken effizient nutzen zu können, benötigen wir diese Matrix in einer speichereffizienten Form. Üblierweise macht man das mit Numpy: <https://numpy.org/doc/stable/user/absolute_beginners.html>

```{code-cell} ipython3
matrix_np = np.array(matrix)
matrix_np
```

```{code-cell} ipython3
from sklearn.metrics.pairwise import cosine_similarity
similarity = cosine_similarity(matrix_np)
```

```{code-cell} ipython3
import seaborn as sns
mask = np.zeros_like(similarity)
mask[np.triu_indices_from(mask)] = True
ax = sns.heatmap(similarity)
```

```{code-cell} ipython3
ax = sns.clustermap(similarity)
# ax.savefig("clustermap.png")
```

```{code-cell} ipython3
files[3]
```

```{code-cell} ipython3

```
